/*
 * wmflame is a windowmaker / afterstep dock/wharf applet for GNU/Linux
 *
 * It draws flames which can be adjusted with little buttons
 *
 * Code based on some of the Window maker programs found on the internet
 *
 * Code based on wmpong, by Dag Wieers <dag@digibel.be>
 *
 * Copyright 1999, Dave Turner (turnerd@reed.edu);
 *
 * This program is distributed under the GPL license. 
 *
 * Thanks to Richard Stallman for the GNU system and the GPL.
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>

#include <X11/xpm.h>
#include <X11/Xlib.h>


#include "../wmgeneral/wmgeneral.h"

#include "wmflame-master.xpm"

#define USLEEP 20000
#define MAX_COLORS 10		/* don't change these */
#define X_MIN 5
#define X_MAX 55.0
#define Y_MIN 5
#define Y_MAX 44.0

#define VERSION  "0.40"

struct xpm_position
  {
    int x, y;
  };
struct xpm_surface
  {
    int x, y, w, h;
  };



struct xpm_position pos[] =
{
  {5, 50},
  {16, 50},
  {27, 50},
  {38, 50},
  {49, 50}
};
struct xpm_surface rel[] =
{
  {1, 70, 11, 9},
  {12, 70, 11, 9},
  {23, 70, 11, 9},
  {34, 70, 11, 9},
  {45, 70, 11, 9}
};
struct xpm_surface prs[] =
{
  {1, 79, 11, 9},
  {12, 79, 11, 9},
  {23, 79, 11, 9},
  {34, 79, 11, 9},
  {45, 79, 11, 9}
};



char seed_line [64];

char wmflame_pm[64][48];

//imported from wmgeneral

extern GC NormalGC;

int bgr [4][64];

/*****************************************************************************/

void init_bgr () {

  int i;

  for (i = 0; i < 64; i ++) {
      bgr [0][i] = i << 2;

  }
  
}
/*****************************************************************************/

void init_seeds () {

  int i;

  for (i = 0; i < 64; i ++) {
    seed_line [i] = (rand() % 64);
  }

}

/*****************************************************************************/

void permute_seeds () {

  int i;

  for (i = 0; i < 64; i ++) {
    seed_line [i] = seed_line [i] - 5 + (rand() % 11);
    if (seed_line [i] < 0) seed_line [i] = 0;
    if (seed_line [i] > 70) seed_line [i] = 70;
  }

}

/*****************************************************************************/
void loaded_permute () {

  int i;

  float j;

  int total = 0, rand_amt;

  /* since consistency is now important (it's a load meter),
     total how much is added, and add on extra (or take some away) 
     if necessary */

  for (i = 0; i < 64; i ++) {
    rand_amt =  (rand() % 21);
    total += rand_amt;

    seed_line [i] = seed_line [i] - 10 + rand_amt;
    if (seed_line [i] < 0) seed_line [i] = 0;
    if (seed_line [i] > 70) seed_line [i] = 70;
  }
  
  if (total != 640) {
    if (640 > total) {
      for (j = 0; j < 64; j += 64.0 / (640 - total)) {
	seed_line [(int)j] ++;
      }
    } else {
      for (j = 0; j < 64; j += 64.0 / (total - 640)) {
	seed_line [(int)j] --;
      }
    }
  }
}
/*****************************************************************************/

void do_frame (float divisor, int random_amt) {

  int x, y;

  for (x = 1; x < 63; x ++) {
    wmflame_pm [x][48] = ((seed_line [x - 1] + seed_line [x] + seed_line [x + 1]) / 2);
  }

  /* this is where the fun goes on.  We sort of "blur up" the flame.  
     There's no color going on yet, just an array of intensity or heat values.
     each pixel gets an average of the point to its right, itself, and the 
     three below it, minus a random little bit.  The addition of the point to 
     the right of the pixel makes the flame appear to move left as though 
     blown by a breeze.  The divisor is always around 5, which is the number 
     of values being averaged.   */

  for (y = 47; y > 0; y --) {
    for (x = 1; x < 62; x ++) {
      wmflame_pm [x][y] = (-(rand() % random_amt) + wmflame_pm [x + 1][y] + wmflame_pm [x][y] + wmflame_pm [x - 1][y + 1] + wmflame_pm [x][y + 1] + wmflame_pm [x + 1][y + 1]) / divisor;
    }
  }

  /* now clean up.  set values less than zero to zero, and values greater 
     than 63 to 63. */

  for (y = 47; y > 1; y --) {
    for (x = 1; x < 62; x ++) {
      if (wmflame_pm [x][y] > 63) {
	wmflame_pm [x][y] = 63;
      } else {
        if (wmflame_pm [x][y] < 0) {
          wmflame_pm [x][y] = 0;
        }
      }
    }
  }

}
/*****************************************************************************/

void loaded_burn (float load) {

  int x, y;

  for (x = 1; x < 63; x ++) {
    wmflame_pm [x][48] = ((seed_line [x - 1] + seed_line [x] + seed_line [x + 1]) / 2);
  }

  /* this is where the fun goes on.  We sort of "blur up" the flame.  
     There's no color going on yet, just an array of intensity or heat values.
     each pixel gets an average of the point to its right, itself, and the 
     three below it, minus a random little bit.  The addition of the point to 
     the right of the pixel makes the flame appear to move left as though 
     blown by a breeze.  The divisor is always around 5, which is the number
     of values being averaged.   */

  for (y = 47; y > 0; y --) {
    for (x = 1; x < 62; x ++) {
      wmflame_pm [x][y] = (wmflame_pm [x][y] + wmflame_pm [x - 1][y + 1] + wmflame_pm [x][y + 1] + wmflame_pm [x + 1][y + 1]) / 
     (4.22 - load / 5.5);
    }
  }

  /* Now clean up.  Set values less than zero to zero, and values greater 
     than 63 to 63. */

  for (y = 47; y > 1; y --) {
    for (x = 1; x < 62; x ++) {
      if (wmflame_pm [x][y] > 63) {
	wmflame_pm [x][y] = 63;
      } else {
        if (wmflame_pm [x][y] < 0) {
          wmflame_pm [x][y] = 0;
        }
      }
    }
  }

}


/*****************************************************************************/

void paint_frame (int cmap) {

  /* I really wish wmgeneral allowed a better way to implement this.
     I have improved speed a little bit by replacing this call to wmgeneral:

        copyXPMArea (wmflame_pm[x][y], 66 + cmap, 1, 1, x, y);

     with this call to Xlib:

	XCopyArea(display, wmgen.pixmap, wmgen.pixmap, NormalGC, wmflame_pm[x][y], 66 + cmap, 1, 1, x, y);

     I hope to further improve performance at some point by not copying, but just blitting a pixel like this:

       XSetForeground (display, NormalGC, bgr [0][wmflame_pm[x][y]]);
       XDrawPoint (display, wmgen.pixmap, NormalGC, x, y);

     This seems to provide a ~30% speed improvement.  However, it requires the bgr array to be created.  It also will probably reduce the portability of wmflame. 

  */

  int x, y;
  for (y = 47; y > 0; y --) {
    for (x = 1; x < 63; x ++) {

      XCopyArea(display, wmgen.pixmap, wmgen.pixmap, NormalGC, wmflame_pm[x][y], 66 + cmap, 1, 1, x, y);


    }
  }
}

float get_load_avg (int which) {

  FILE *fp;
  float ftmp = 0;
  int i;

  if( (fp = fopen("/proc/loadavg", "r")) != NULL) {
    for (i = 0; i < which; i ++) {
      fscanf(fp, "%f", &ftmp);
    }
  }

  fclose (fp);
  return ftmp;
}
/*****************************************************************************/

void print_usage () {

  printf ("\nWmFlame %s  \nBy Dave Turner (turnerd@reed.edu)\n", VERSION);
  printf ("Usage:\n\n");
  printf ("      -h        This screen\n");
  printf ("      -lN       Use as load meter.\n");
  printf ("        N = 1:   Load average over 1 minute\n");
  printf ("        N = 2:   Load average over 5 minutes\n");
  printf ("        N = 3:   Load average over 15 minutes\n\n");
  printf ("      -v        Print version number\n");
  printf ("      -m N      Load max is N.  You usually want N to be equal to the number\n");
  printf ("                of CPUs you have.\n\n");
}

int main (int argc, char *argv[]) {
  XEvent Event;

  char wmflame_mask_bits[64*64];
  int wmflame_mask_width = 64;
  int wmflame_mask_height = 64;

  int random_amt = 22;   // these two values were obtained through trial
  float divisor = 4.79;  // and error.  Change them here, or with the 
                         // little buttons.;			      
  int cmap = 0;
  
  int i;

  int paused = 0;  /* if non-zero, doesn't draw anything */

  int load_meter = 0; /*is it a load meter?  if so, which average?*/
  float maximum = 1; /*if it's a load meter, this determines max load */

  float load = 0;


  //This function only needs to be called if we're using the XDrawPoint optimization
  //As this isn't complete, don't do it.
  //init_bgr ();

  for (i=1; i< argc; i++) {
    char *arg = argv[i];
    
    if (*arg=='-') {
      switch (arg[1]) {

      case 'v' :
	printf ("%s\n", VERSION);
	exit (0);
        break;
      case 'l' :

	switch (arg[2]) {
	case '1':
	  load_meter = 1;
	  break;
	case '2':
	  load_meter = 2;
	  break;
	case '3':
	  load_meter = 3;
	  break;
	default:
	  load_meter = 1;
	  break;
	}
	break;

      case 'm':
	i ++;
	maximum = atof (argv [i]);
	break;

      default:
	print_usage();
	exit (0);
	break;
      }
    }
  }

  srand (time (NULL));

  init_seeds ();

  createXBMfromXPM (wmflame_mask_bits, wmflame_master_xpm, wmflame_mask_width, wmflame_mask_height);

  openXwindow (argc, argv, wmflame_master_xpm, wmflame_mask_bits, wmflame_mask_width, wmflame_mask_height);
  
  for (i = 0; i < 5; i++)
    AddMouseRegion (i, pos[i].x, pos[i].y, pos[i].x + prs[i].w, pos[i].y + prs[i].h);
  
  AddMouseRegion (5, 0, 0, 64, 47);

  while (1) {

    if ( !paused ) {
      if (!load_meter) {
	
        permute_seeds ();
	do_frame (divisor, random_amt);
	
      } else {
	load = get_load_avg (load_meter);
	if (load > maximum) load = maximum;
	
	loaded_permute ();
	loaded_burn (load / maximum);
      }
      paint_frame (cmap);
      RedrawWindow ();    
    } else {
      usleep (USLEEP);
    }

    while (XPending (display)) {
      XNextEvent (display, &Event);
      switch (Event.type) {

      case Expose:
	RedrawWindow ();
	break;
      case DestroyNotify:
	XCloseDisplay (display);
	exit (0);
	break;
      case ButtonPress:
        if ( Event.xbutton.button == 2 ) {
          paused = !paused;
          break;
        }

	i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
	if (i < 5) copyXPMArea (prs[i].x, prs[i].y, prs[i].w, prs[i].h, pos[i].x, pos[i].y);
        switch (i) {
	case 0:
	  divisor -= .01;
	  if (divisor <= 0) divisor = .01;
	  break;
	case 1:
	  divisor += .01;
	  break;
	case 2:
	  random_amt += 1;
	  break;
	case 3:
	  random_amt -= 1;
	  if (random_amt < 1) random_amt = 1;
	  break;
	case 4:
	  init_seeds ();
	  break;
	  
	case 5:
	  
	  cmap = (cmap + 1) % 3;
	  random_amt = 22; 
	  divisor = 4.79;
                         
	  break;
	}

	break;

      case ButtonRelease:
	i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
	if (i < 5) copyXPMArea (rel[i].x, rel[i].y, rel[i].w, rel[i].h, pos[i].x, pos[i].y);
	break;
      }
    }
    usleep (USLEEP);
  }

  return 0;

}


